#!/usr/bin/perl
#
# Extract supported MIME type per Debian package as reported by the
# packages .desktop files.  Work on a Debian mirror and extract the
# debs to peek at the .desktop files.

use warnings;
use strict;

my $mirrordir = "~ftp/debian";
my $suite = $ARGV[1] || 'stable';
my $arch = $ARGV[2] || 'i386';

my %pkgs;
open(my $fh, "zcat $mirrordir/dists/$suite/Contents-$arch.gz |");
my $skip = 1;
while (<$fh>) {
    if (/^FILE\s+LOCATION/) {
        $skip = 0;
        next;
    }
    next if $skip;
    chomp;
    my ($path, $pkg) = split(/\s+/, $_, 2);
#    print STDERR "$path $pkg\n";
    if (
       $path =~ m%^usr/share/applications/.*\.desktop$% ||
       $path =~ m%^lib/modules/.*\.ko$% ||
       $path =~ m%^lib/firmware/.+$% ||
       $path =~ m%^lib/udev/rules.d/.+$%
       $path =~ m%^usr/share/modaliases/.+$%
       ) {
        my @f = split(/\//, $pkg);
#        print "$f[-1] - $path\n";
        $pkgs{$f[-1]} = 1;
    }
}
close($fh);

my %pkgmap;
for my $section (qw(main contrib non-free)) {
    open(my $fh, "zcat $mirrordir/dists/$suite/$section/binary-$arch/Packages.gz |");
    while (<$fh>) {
        if (m/^Filename: (.+)$/) {
            my $path = $1;
            my ($binarypkg) = $path =~ m%.+/([^/_]+)_.*$%;
            $pkgmap{$binarypkg} = $path;
#            print "$binarypkg\n";
        }
    }
}

for my $pkg (sort keys %pkgs) {
    my $path = $pkgmap{$pkg};
#    print $pkg, $path, "\n";
    unless ($path) {
        print STDERR "error: no path for package $pkg\n";
        next;
    }
    my $tmpdir = `mktemp -d`;
    chomp $tmpdir;
    open(my $fh, "dpkg-deb -D --vextract $mirrordir/$path $tmpdir |") || warn "unable to find $pkg";
    my %pkgtypes;
    my %pkgaliases;
    my %pkgfirmware;
#    print "Fetching from dpkg-deb $path\n";
    while (<$fh>) {
#        print;
        chomp;
        my $debline = $_;
        if ($debline =~ m%^.(/usr/share/applications/.*\.desktop)%) {
            my $desktopfile = $1;
            my $dfpath = "$tmpdir/$desktopfile";
            if ( -r $dfpath ) {
                open(my $dfh, '<', $dfpath ) ||
                    warn "unable to open desktop file $desktopfile from $path";
                while (<$dfh>) {
#                print " D: ", $_;
                    chomp;
                    if (m/MimeType=(.+)/ && defined $1) {
                        for my $mimetype (split(/;/, $1)) {
                            $pkgtypes{$mimetype} = 1;
#                        print "$pkg $mimetype\n";
                        }
                    }
                }
                close($dfh);
            }
        }
        if ($debline =~ m%^.(/lib/modules/.*\.ko)%) {
            my $kopath = "$tmpdir/$1";
            open(my $sfh, "strings $kopath |");
            while (<$sfh>) {
                if (m/^alias=(.+)$/ && defined $1) {
                    my $modalias = $1;
#                   print STDERR "A: $modalias\n";
                    $pkgaliases{$modalias} = 1 if $modalias =~ m/:/;
                }
                if (m/^firmware=(.+)$/ && defined $1) {
#                   print STDERR "F: $1\n";
                }
            }
        }
        if ($pkg !~ m/^linux-image/
            && $debline =~ m%^./lib/firmware/(.+)$%
            ) {
            if ( ! -d "$tmpdir/lib/firmware/$1" ) {
                $pkgfirmware{$1} = 1;
            }
        }
        # Try to guess modalias entries based on udev rules
        if  ($debline =~ m%^./(lib/udev/rules.d/.+)$%) {
            my $rulepath = "$tmpdir/$1";
            open(my $rfh, '<', $rulepath) || next;
            my $line = "";
            my $subsystem = "";
            my $action;
            my $idvendor;
            my $idproduct;
            while (<$rfh>) {
                if (/\\$/) {
                    chomp;
                    $_ =~ s/\\$//;
                    $line .= $_;
                    next;
                }
                $_ = $line . $_;
                $subsystem = $1 if (m/SUBSYSTEM=="([^"]+)"/);
#                $action = $1    if (m/ACTION=="([^"]+)"/);
                $idvendor = $1  if (m/ATTR{idVendor}=="([^"]+)"/);
                $idproduct = $1 if (m/ATTR{idProduct}=="([^"]+)"/);
                $line = "";
                $subsystem = "usb" if "usb_device" eq $subsystem;
                if ($subsystem && $idvendor && $idproduct) {
#                    print $_;
                    my $modalias = sprintf("%s:v%sp%sd*",
                                           $subsystem,
                                           uc($idvendor),
                                           uc($idproduct));
#                    print "$rulepath $modalias\n";
                    $pkgaliases{$modalias} = $rulepath;
                    $idvendor = "";
                    $idproduct = "";
                }
            }
            close($rfh);
        }
    }
    close($fh);
    my $headers = "";
    if (0 < scalar keys %pkgtypes) {
        $headers .= "MimeType: " . join(";", sort keys %pkgtypes) . "\n";
    }
    if (0 < scalar keys %pkgaliases) {
        $headers .= "Modaliases: $pkg(" . join(", ", sort keys %pkgaliases) . ")\n";
    }
    if (0 < scalar keys %pkgfirmware) {
        $headers .= "Firmware: " . join(", ", sort keys %pkgfirmware) . "\n";
    }
    print "Package: $pkg\n$headers\n" if $headers;
    system("rm -rf $tmpdir");
}
